import numpy.typing
from typing import Tuple, Sequence, Union
from pdme.measurement.dot_measure import DotRangeMeasurement
from pdme.measurement.dot_pair_measure import DotPairRangeMeasurement


DotInput = Tuple[numpy.typing.ArrayLike, float]
DotPairInput = Tuple[numpy.typing.ArrayLike, numpy.typing.ArrayLike, float]


def dot_inputs_to_array(dot_inputs: Sequence[DotInput]) -> numpy.ndarray:
	return numpy.array([numpy.append(numpy.array(input[0]), input[1]) for input in dot_inputs])


def dot_pair_inputs_to_array(pair_inputs: Sequence[DotPairInput]) -> numpy.ndarray:
	return numpy.array([[numpy.append(numpy.array(input[0]), input[2]), numpy.append(numpy.array(input[1]), input[2])] for input in pair_inputs])


def dot_range_measurements_low_high_arrays(dot_range_measurements: Union[Sequence[DotRangeMeasurement], Sequence[DotPairRangeMeasurement]]) -> Tuple[numpy.ndarray, numpy.ndarray]:
	lows = [measurement.v_low for measurement in dot_range_measurements]
	highs = [measurement.v_high for measurement in dot_range_measurements]
	return (numpy.array(lows), numpy.array(highs))



from pdme.model.model import Model, Discretisation
from pdme.model.fixed_z_plane_model import FixedZPlaneModel
from pdme.model.unrestricted_model import UnrestrictedModel
from pdme.model.fixed_dipole_model import FixedDipoleModel
from pdme.model.fixed_magnitude_model import FixedMagnitudeModel

__all__ = ["Model", "Discretisation", "FixedZPlaneModel", "UnrestrictedModel", "FixedDipoleModel", "FixedMagnitudeModel"]



from dataclasses import dataclass
import numpy
import numpy.typing


@dataclass
class DotPairMeasurement():
	'''
	Representation of a dot measuring oscillating dipoles.

	Parameters
	----------
	v : float
		The voltage measured at the dot.
	r1 : numpy.ndarray
		The position of the first dot.
	r2 : numpy.ndarray
		The position of the second dot.
	f : float
		The measurement frequency.
	'''
	v: float
	r1: numpy.ndarray
	r2: numpy.ndarray
	f: float

	def __post_init__(self) -> None:
		self.r1 = numpy.array(self.r1)
		self.r2 = numpy.array(self.r2)


@dataclass
class DotPairRangeMeasurement():
	'''
	Representation of a dot measuring oscillating dipoles.

	Parameters
	----------
	v_low : float
		The lower range of voltage measured at the dot.
	v_high : float
		The upper range of voltage measured at the dot.
	r1 : numpy.ndarray
		The position of the first dot.
	r2 : numpy.ndarray
		The position of the second dot.
	f : float
		The measurement frequency.
	'''
	v_low: float
	v_high: float
	r1: numpy.ndarray
	r2: numpy.ndarray
	f: float

	def __post_init__(self) -> None:
		self.r1 = numpy.array(self.r1)
		self.r2 = numpy.array(self.r2)
		if self.v_low > self.v_high:
			self.v_low, self.v_high = self.v_high, self.v_low



import numpy
import scipy.optimize
from typing import Callable, Sequence, Tuple, List
from pdme.measurement import DotMeasurement, OscillatingDipoleArrangement, OscillatingDipole
import pdme.util
import logging


_logger = logging.getLogger(__name__)


class Model():
	"""
	Interface for models.
	"""

	def point_length(self) -> int:
		raise NotImplementedError

	def n(self) -> int:
		raise NotImplementedError

	def v_for_point_at_dot(self, dot: DotMeasurement, pt: numpy.ndarray) -> float:
		raise NotImplementedError

	def get_dipoles(self, frequency: float) -> OscillatingDipoleArrangement:
		raise NotImplementedError

	def get_n_single_dipoles(self, n: int, max_frequency: float) -> numpy.ndarray:
		raise NotImplementedError

	def solution_single_dipole(self, pt: numpy.ndarray) -> OscillatingDipole:
		raise NotImplementedError

	def solution_as_dipoles(self, pts: numpy.ndarray) -> List[OscillatingDipole]:
		pt_length = self.point_length()
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]
		return [self.solution_single_dipole(pt) for pt in chunked_pts]

	def cost_for_dot(self, dot: DotMeasurement, pts: numpy.ndarray) -> float:
		# creates numpy.ndarrays in groups of self.point_length().
		# Will throw problems for irregular points, but that's okay for now.
		pt_length = self.point_length()
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]
		return sum(self.v_for_point_at_dot(dot, pt) for pt in chunked_pts) - dot.v

	def costs(self, dots: Sequence[DotMeasurement]) -> Callable[[numpy.ndarray], numpy.ndarray]:
		'''
		Returns a function that returns the cost for the given list of DotMeasurements for a particular model-dependent phase space point.
		Default implementation assumes a single dot cost from which to build the list.

		Parameters
		----------
		dots: A list of dot measurements to use to find the cost functions.

		Returns
		----------
		Returns the model's cost function.
		'''
		_logger.debug(f"Constructing costs for dots: {dots}")

		def costs_to_return(pts: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([self.cost_for_dot(dot, pts) for dot in dots])

		return costs_to_return

	def jac_for_point_at_dot(self, dot: DotMeasurement, pt: numpy.ndarray) -> numpy.ndarray:
		raise NotImplementedError

	def jac_for_dot(self, dot: DotMeasurement, pts: numpy.ndarray) -> numpy.ndarray:
		# creates numpy.ndarrays in groups of self.point_length().
		# Will throw problems for irregular points, but that's okay for now.
		pt_length = self.point_length()
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]
		return numpy.append([], [self.jac_for_point_at_dot(dot, pt) for pt in chunked_pts])

	def jac(self, dots: Sequence[DotMeasurement]) -> Callable[[numpy.ndarray], numpy.ndarray]:
		'''
		Returns a function that returns the cost function's Jacobian for the given list of DotMeasurements for a particular model-dependent phase space point.
		Default implementation assumes a single dot jacobian from which to build the list.

		Parameters
		----------
		dots: A list of dot measurements to use to find the cost functions and their Jacobian.

		Returns
		----------
		Returns the model's cost function's Jacobian.
		'''
		def jac_to_return(pts: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([self.jac_for_dot(dot, pts) for dot in dots])

		return jac_to_return

	def solve(self, dots: Sequence[DotMeasurement], initial_pt: numpy.ndarray = None, bounds=(-numpy.inf, numpy.inf)) -> scipy.optimize.OptimizeResult:
		if initial_pt is None:
			initial = numpy.tile(.1, self.n() * self.point_length())
		else:
			if len(initial_pt) != self.point_length():
				raise ValueError(f"The initial point {initial_pt} does not have the model's expected length: {self.point_length()}")
			initial = numpy.tile(initial_pt, self.n())

		result = scipy.optimize.least_squares(self.costs(dots), initial, jac=self.jac(dots), ftol=1e-15, gtol=3e-16, xtol=None, bounds=bounds)
		result.normalised_x = pdme.util.normalise_point_list(result.x, self.point_length())
		return result


class Discretisation():
	def bounds(self, index: Tuple[float, ...]) -> Tuple:
		raise NotImplementedError

	def all_indices(self) -> numpy.ndindex:
		raise NotImplementedError

	def solve_for_index(self, dots: Sequence[DotMeasurement], index: Tuple) -> scipy.optimize.OptimizeResult:
		raise NotImplementedError

	def get_model(self) -> Model:
		raise NotImplementedError



import numpy
from dataclasses import dataclass
from typing import Sequence, Tuple
import scipy.optimize
from pdme.model.model import Model, Discretisation
from pdme.measurement import DotMeasurement, OscillatingDipoleArrangement, OscillatingDipole


class UnrestrictedModel(Model):
	'''
	Model of oscillating dipoles with no restrictions.
	Additionally, each dipole is assumed to be orientated in the plus or minus z direction.

	Parameters
	----------
	n : int
		The number of dipoles to assume.
	'''
	def __init__(self, xmin: float, xmax: float, ymin: float, ymax: float, zmin: float, zmax: float, max_p: float, n: int) -> None:
		self.xmin = xmin
		self.xmax = xmax
		self.ymin = ymin
		self.ymax = ymax
		self.zmin = zmin
		self.zmax = zmax
		self.max_p = max_p
		self._n = n
		self.rng = numpy.random.default_rng()

	def __repr__(self) -> str:
		return f'UnrestrictedModel({self.xmin}, {self.xmax}, {self.ymin}, {self.ymax}, {self.zmin}, {self.zmax}, {self.max_p}, {self.n()})'

	def point_length(self) -> int:
		'''
			Dipole is unconstrained in this model.
			All seven degrees of freedom: (px, py, pz, sx, sy, sz, w).
		'''
		return 7

	def n(self) -> int:
		return self._n

	def v_for_point_at_dot(self, dot: DotMeasurement, pt: numpy.ndarray) -> float:
		p = pt[0:3]
		s = pt[3:6]
		w = pt[6]

		diff = dot.r - s
		alpha = p.dot(diff) / (numpy.linalg.norm(diff)**3)
		b = (1 / numpy.pi) * (w / (w**2 + dot.f**2))
		return alpha**2 * b

	def jac_for_point_at_dot(self, dot: DotMeasurement, pt: numpy.ndarray) -> numpy.ndarray:
		p = pt[0:3]
		s = pt[3:6]
		w = pt[6]

		diff = dot.r - s
		alpha = p.dot(diff) / (numpy.linalg.norm(diff)**3)
		b = (1 / numpy.pi) * (w / (w**2 + dot.f**2))

		p_divs = 2 * alpha * diff / (numpy.linalg.norm(diff)**3) * b

		r_divs = (-p / (numpy.linalg.norm(diff)**3) + 3 * p.dot(diff) * diff / (numpy.linalg.norm(diff)**5)) * 2 * alpha * b

		f2 = dot.f**2
		w2 = w**2

		w_div = alpha**2 * (1 / numpy.pi) * ((f2 - w2) / ((f2 + w2)**2))

		return numpy.concatenate((p_divs, r_divs, w_div), axis=None)

	def get_dipoles(self, frequency: float) -> OscillatingDipoleArrangement:
		theta = numpy.arccos(self.rng.uniform(-1, 1))
		phi = self.rng.uniform(0, 2 * numpy.pi)
		p = self.rng.uniform(0, self.max_p)
		px = p * numpy.sin(theta) * numpy.cos(phi)
		py = p * numpy.sin(theta) * numpy.sin(phi)
		pz = p * numpy.cos(theta)
		s_pts = numpy.array((self.rng.uniform(self.xmin, self.xmax), self.rng.uniform(self.ymin, self.ymax), self.rng.uniform(self.zmin, self.zmax)))
		return OscillatingDipoleArrangement([OscillatingDipole(numpy.array([px, py, pz]), s_pts, frequency)])


@dataclass
class UnrestrictedDiscretisation(Discretisation):
	'''
	Representation of a discretisation of a UnrestrictedModel.
	Also captures a rough maximum value of dipole.

	Parameters
	----------
	model : UnrestrictedModel
		The parent model of the discretisation.
	num_px: int
		The number of partitions of the px.
	num_py: int
		The number of partitions of the py.
	num_pz: int
		The number of partitions of pz.
	num_x : int
		The number of partitions of the x axis.
	num_y : int
		The number of partitions of the y axis.
	num_z : int
		The number of partitions of the z axis.
	max_p : int
		The maximum p coordinate in any direction.
	'''
	model: UnrestrictedModel
	num_px: int
	num_py: int
	num_pz: int
	num_x: int
	num_y: int
	num_z: int

	def __post_init__(self):
		self.max_p = self.model.max_p
		self.cell_count = self.num_x * self.num_y * self.num_z
		self.x_step = (self.model.xmax - self.model.xmin) / self.num_x
		self.y_step = (self.model.ymax - self.model.ymin) / self.num_y
		self.z_step = (self.model.zmax - self.model.zmin) / self.num_z
		self.px_step = 2 * self.max_p / self.num_px
		self.py_step = 2 * self.max_p / self.num_py
		self.pz_step = 2 * self.max_p / self.num_pz

	def bounds(self, index: Tuple[float, ...]) -> Tuple:
		pxi, pyi, pzi, xi, yi, zi = index

		# For this model, a point is (px, py, pz, sx, sx, sy, w).
		# We want to keep w unbounded, restrict sx, sy, sz, px and py based on step.
		return (
			[
				pxi * self.px_step - self.max_p, pyi * self.py_step - self.max_p, pzi * self.pz_step - self.max_p,
				xi * self.x_step + self.model.xmin, yi * self.y_step + self.model.ymin, zi * self.z_step + self.model.zmin,
				-numpy.inf
			],
			[
				(pxi + 1) * self.px_step - self.max_p, (pyi + 1) * self.py_step - self.max_p, (pzi + 1) * self.pz_step - self.max_p,
				(xi + 1) * self.x_step + self.model.xmin, (yi + 1) * self.y_step + self.model.ymin, (zi + 1) * self.z_step + self.model.zmin,
				numpy.inf
			]
		)

	def all_indices(self) -> numpy.ndindex:
		# see https://github.com/numpy/numpy/issues/20706 for why this is a mypy problem.
		return numpy.ndindex((self.num_px, self.num_py, self.num_pz, self.num_x, self.num_y, self.num_z))  # type:ignore

	def solve_for_index(self, dots: Sequence[DotMeasurement], index: Tuple[float, ...]) -> scipy.optimize.OptimizeResult:
		bounds = self.bounds(index)
		px_mean = (bounds[0][0] + bounds[1][0]) / 2
		py_mean = (bounds[0][1] + bounds[1][1]) / 2
		pz_mean = (bounds[0][2] + bounds[1][2]) / 2
		sx_mean = (bounds[0][3] + bounds[1][3]) / 2
		sy_mean = (bounds[0][4] + bounds[1][4]) / 2
		sz_mean = (bounds[0][5] + bounds[1][5]) / 2
		return self.model.solve(dots, initial_pt=numpy.array([px_mean, py_mean, pz_mean, sx_mean, sy_mean, sz_mean, .1]), bounds=bounds)



import numpy
from dataclasses import dataclass
from typing import Tuple, Sequence
import scipy.optimize

from pdme.model.model import Model
from pdme.measurement import DotMeasurement


class FixedZPlaneModel(Model):
	'''
	Model of oscillating dipoles constrained to lie within a plane.
	Additionally, each dipole is assumed to be orientated in the plus or minus z direction.

	Parameters
	----------
	z : float
		The z position of the plane where dipoles are constrained to lie.
	xmin : float
		The minimum x value for dipoles.
	xmax : float
		The maximum x value for dipoles.
	ymin : float
		The minimum y value for dipoles.
	ymax : float
		The maximum y value for dipoles.
	n : int
		The number of dipoles to assume.
	'''
	def __init__(self, z: float, xmin: float, xmax: float, ymin: float, ymax: float, n: int) -> None:
		self.z = z
		self.xmin = xmin
		self.xmax = xmax
		self.ymin = ymin
		self.ymax = ymax
		self._n = n

	def __repr__(self) -> str:
		return f'FixedZPlaneModel({self.z}, {self.xmin}, {self.xmax}, {self.ymin}, {self.ymax}, {self.n()})'

	def point_length(self) -> int:
		'''
			Dipole is constrained in this model to have (px, py, pz) = (0, 0, pz) and (sx, sy, sz) = (sx, sy, self.z).
			With some frequency w, there are four degrees of freedom: (pz, sx, sy, w).
		'''
		return 4

	def n(self) -> int:
		return self._n

	def v_for_point_at_dot(self, dot: DotMeasurement, pt: numpy.ndarray) -> float:
		p = numpy.array([0, 0, pt[0]])
		s = numpy.array([pt[1], pt[2], self.z])
		w = pt[3]

		diff = dot.r - s
		alpha = p.dot(diff) / (numpy.linalg.norm(diff)**3)
		b = (1 / numpy.pi) * (w / (w**2 + dot.f**2))
		return alpha**2 * b

	def jac_for_point_at_dot(self, dot: DotMeasurement, pt: numpy.ndarray) -> numpy.ndarray:
		p = numpy.array([0, 0, pt[0]])
		s = numpy.array([pt[1], pt[2], self.z])
		w = pt[3]

		diff = dot.r - s
		alpha = p.dot(diff) / (numpy.linalg.norm(diff)**3)
		b = (1 / numpy.pi) * (w / (w**2 + dot.f**2))

		p_divs = 2 * alpha * diff[2] / (numpy.linalg.norm(diff)**3) * b  # only need the z component.

		r_divs = (-p[0:2] / (numpy.linalg.norm(diff)**3) + 3 * p.dot(diff) * diff[0:2] / (numpy.linalg.norm(diff)**5)) * 2 * alpha * b

		f2 = dot.f**2
		w2 = w**2

		w_div = alpha**2 * (1 / numpy.pi) * ((f2 - w2) / ((f2 + w2)**2))

		return numpy.concatenate((p_divs, r_divs, w_div), axis=None)


@dataclass
class FixedZPlaneDiscretisation():
	'''
	Representation of a discretisation of a FixedZPlaneModel.
	Also captures a rough maximum value of dipole.

	Parameters
	----------
	model : FixedZPlaneModel
		The parent model of the discretisation.
	num_pz: int
		The number of partitions of pz.
	num_x : int
		The number of partitions of the x axis.
	num_y : int
		The number of partitions of the y axis.
	'''
	model: FixedZPlaneModel
	num_pz: int
	num_x: int
	num_y: int
	max_pz: int

	def __post_init__(self):
		self.cell_count = self.num_x * self.num_y
		self.pz_step = (2 * self.max_pz) / self.num_pz
		self.x_step = (self.model.xmax - self.model.xmin) / self.num_x
		self.y_step = (self.model.ymax - self.model.ymin) / self.num_y

	def bounds(self, index: Tuple[float, float, float]) -> Tuple[numpy.ndarray, numpy.ndarray]:
		pzi, xi, yi = index

		# For this model, a point is (pz, sx, sy, w).
		# We want to keep w bounded, and restrict pz, sx and sy based on step.
		return (
			numpy.array((pzi * self.pz_step - self.max_pz, xi * self.x_step + self.model.xmin, yi * self.y_step + self.model.ymin, -numpy.inf)),
			numpy.array(((pzi + 1) * self.pz_step - self.max_pz, (xi + 1) * self.x_step + self.model.xmin, (yi + 1) * self.y_step + self.model.ymin, numpy.inf))
		)

	def all_indices(self) -> numpy.ndindex:
		# see https://github.com/numpy/numpy/issues/20706 for why this is a mypy problem.
		return numpy.ndindex((self.num_pz, self.num_x, self.num_y))  # type:ignore

	def solve_for_index(self, dots: Sequence[DotMeasurement], index: Tuple[float, float, float]) -> scipy.optimize.OptimizeResult:
		bounds = self.bounds(index)
		pz_mean = (bounds[0][0] + bounds[1][0]) / 2
		sx_mean = (bounds[0][1] + bounds[1][1]) / 2
		sy_mean = (bounds[0][2] + bounds[1][2]) / 2
		# I don't care about the typing here at the moment.
		return self.model.solve(dots, initial_pt=numpy.array((pz_mean, sx_mean, sy_mean, .1)), bounds=bounds)  # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pdme', 'pdme.inputs', 'pdme.measurement', 'pdme.model', 'pdme.util']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.1,<2.0.0', 'scipy>=1.8,<1.9']

setup_kwargs = {
    'name': 'pdme',
    'version': '0.6.1',
    'description': 'Python dipole model evaluator',
    'long_description': '# pdme - the python dipole model evaluator\n\n[![Conventional Commits](https://img.shields.io/badge/Conventional%20Commits-1.0.0-green.svg?style=flat-square)](https://conventionalcommits.org)\n[![PyPI](https://img.shields.io/pypi/v/pdme?style=flat-square)](https://pypi.org/project/pdme/)\n[![Jenkins](https://img.shields.io/jenkins/build?jobUrl=https%3A%2F%2Fjenkins.deepak.science%2Fjob%2Fgitea-physics%2Fjob%2Fpdme%2Fjob%2Fmaster&style=flat-square)](https://jenkins.deepak.science/job/gitea-physics/job/pdme/job/master/)\n![Jenkins tests](https://img.shields.io/jenkins/tests?compact_message&jobUrl=https%3A%2F%2Fjenkins.deepak.science%2Fjob%2Fgitea-physics%2Fjob%2Fpdme%2Fjob%2Fmaster%2F&style=flat-square)\n![Jenkins Coverage](https://img.shields.io/jenkins/coverage/cobertura?jobUrl=https%3A%2F%2Fjenkins.deepak.science%2Fjob%2Fgitea-physics%2Fjob%2Fpdme%2Fjob%2Fmaster%2F&style=flat-square)\n![Maintenance](https://img.shields.io/maintenance/yes/2022?style=flat-square)\n\nThis repo has library code for evaluating dipole models.\n\n## Getting started\n\n`poetry install` to start locally\n\nCommit using [Conventional Commits](https://www.conventionalcommits.org/en/v1.0.0/), and when commits are on master, release with `doo release`.\n',
    'author': 'Deepak',
    'author_email': 'dmallubhotla+github@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import numpy
import logging


_logger = logging.getLogger(__name__)


def fast_vs_for_dipoles(dot_inputs: numpy.ndarray, dipoles: numpy.ndarray) -> numpy.ndarray:
	'''
		No error correction here baby.
	'''
	ps = dipoles[:, 0:3]
	ss = dipoles[:, 3:6]
	ws = dipoles[:, 6]

	_logger.debug(f"ps: {ps}")
	_logger.debug(f"ss: {ss}")
	_logger.debug(f"ws: {ws}")

	rs = dot_inputs[:, 0:3]
	fs = dot_inputs[:, 3]

	diffses = rs - ss[:, None]

	_logger.debug(f"diffses: {diffses}")
	norms = numpy.linalg.norm(diffses, axis=2)**3
	_logger.debug(f"norms: {norms}")
	ases = (numpy.einsum('...ji, ...i', diffses, ps) / norms)**2
	_logger.debug(f"ases: {ases}")

	bses = (1 / numpy.pi) * (ws[:, None] / (fs**2 + ws[:, None]**2))
	_logger.debug(f"bses: {bses}")
	return ases * bses


def between(a: numpy.ndarray, low: numpy.ndarray, high: numpy.ndarray) -> numpy.ndarray:
	'''
		Intended specifically for the case where a is a list of arrays, and each array must be between the single array low and high, but without error checking.
	'''
	return numpy.all(numpy.logical_and(low < a, high > a), axis=1)



import numpy
import operator
import logging


# flips px, py, pz
SIGN_ARRAY_7 = numpy.array((-1, -1, -1, 1, 1, 1, 1))
SIGN_ARRAY_4 = numpy.array((-1, 1, 1, 1))


_logger = logging.getLogger(__name__)


def flip_chunk_to_positive_px(pt: numpy.ndarray) -> numpy.ndarray:
	if pt[0] > 0:
		return pt
	else:
		# godawful hack.
		if len(pt) == 7:
			return SIGN_ARRAY_7 * pt
		elif len(pt) == 4:
			return SIGN_ARRAY_4 * pt
		else:
			_logger.warning(f"Could not normalise pt: {pt}. Returning as is...")
			return pt


def normalise_point_list(pts: numpy.ndarray, pt_length) -> numpy.ndarray:
	chunked_pts = [flip_chunk_to_positive_px(pts[i: i + pt_length]) for i in range(0, len(pts), pt_length)]
	range_to_length = list(range(pt_length))
	rotated_range = range_to_length[pt_length - 1:] + range_to_length[0:pt_length - 1]
	return numpy.concatenate(sorted(chunked_pts, key=lambda x: tuple(round(val, 3) for val in operator.itemgetter(*rotated_range)(x))), axis=None)



from pdme.measurement.dot_measure import DotMeasurement, DotRangeMeasurement
from pdme.measurement.dot_pair_measure import DotPairMeasurement, DotPairRangeMeasurement
from pdme.measurement.oscillating_dipole import OscillatingDipole, OscillatingDipoleArrangement
from pdme.measurement.input_types import DotInput, DotPairInput


__all__ = ['DotMeasurement', 'DotRangeMeasurement', 'DotPairMeasurement', 'DotPairRangeMeasurement', 'OscillatingDipole', 'OscillatingDipoleArrangement', 'DotInput', 'DotPairInput']
